#include "project.h"

#include <string_view>
#include <exception>

using json = nlohmann::json;

constexpr std::string_view entry("mason.json");

std::map<std::string, Dependency::Type> types({
    {"simple", Dependency::Type::simple},
    {"none", Dependency::Type::simple},
    {"mason", Dependency::Type::mason},
    {"auto", Dependency::Type::automatic},
    {"automatic", Dependency::Type::automatic}
});

Project::Project(const std::filesystem::path& location):
    location(location),
    status(Status::unknown),
    name(),
    logStorage(),
    dependencies()
{}

bool Project::read() {
    if (status == Status::read)
        return true;

    std::filesystem::path entryPointPath;
    try {
        entryPointPath = std::filesystem::canonical(location) / entry;
    } catch (const std::exception& e) {
        log(e.what());
        status = Status::error;
        return false;
    }

    std::ifstream file(entryPointPath);
    if (!file.is_open()) {
        log("couldn't open " + std::string(entryPointPath));
        status = Status::error;
        return false;
    }

    json data;
    try {
        data = json::parse(file);
    } catch (const json::exception& e) {
        log(e.what());
        status = Status::error;
        return false;
    }

    try {
        parse(data);
    } catch (const std::exception& e) {
        log(e.what());
        status = Status::error;
        return false;
    }

    return true;
}

void Project::parse(const json& data) {
    const json& nm = data.at("name");
    if (!nm.is_string())
        throw 1;

    name = nm;

    const json& dpd = data.at("dependencies");
    if (!dpd.is_array())
        throw 1;

    for (const json& dep : dpd) {
        switch (dep.type()) {
            case json::value_t::string:
                createDepencencyFromString(dep);
                break;
            case json::value_t::object:
                createDepencencyFromObject(dep);
                break;
            default:
                throw 1;
        }
    }
}

void Project::createDepencencyFromString(const std::string& entry) {
    dependencies.emplace(entry, entry);
}

void Project::createDepencencyFromObject(const nlohmann::json& entry) {
    std::string url;
    json::const_iterator itr = entry.find("url");
    if (itr == entry.end()) {
        itr = entry.find("path");
        if (itr == entry.end())
            throw 1;
    }
    if (!itr->is_string())
        throw 1;

    url = *itr;

    std::optional<std::string> name = std::nullopt;
    itr = entry.find("name");
    if (itr != entry.end() && itr->is_string())
        name = *itr;

    std::optional<std::string> version = std::nullopt;
    itr = entry.find("version");
    if (itr != entry.end() && itr->is_string())
        version = *itr;

    Dependency::Type type = Dependency::Type::automatic;
    itr = entry.find("type");
    if (itr != entry.end() && itr->is_string()) {
        std::map<std::string, Dependency::Type>::const_iterator titr = types.find(*itr);
        if (titr != types.end())
            type = titr->second;
    }

    dependencies.emplace(url, Dependency{url, type, name, version});
}

uint32_t Project::dependenciesCount() const {
    return dependencies.size();
}


void Project::log(const std::string& message) const {
    logStorage.emplace_back(message);
}

Project::Log Project::getLog() const {
    return logStorage;
}

Project::Status Project::getStatus() const {
    return status;
}

std::string Project::getName() const {
    return name;
}
